!! module to implent the functions related to the fitness function evaluation.
!! The actual cost function is problem dependent and is implemented in a
!! separate module. Include everyting related to your function evaluation into
!! a module and replace the statement "use ga_Mauro_function, only: func" with 
!! the name of your module. This module will comunicate with your module only
!! through function func which should have the following interface:
!!					call func(npopsiz,nparam,j,funcval,parent)
!! where all parameters are input, except funcval which is the fitness value 
!! for the individual "j" in the "parent" population


module ga_fitness
	use ga_aux
	use ga_Mauro_function, only: func

	implicit none

	contains

	!------------------------------------------------------------------------
	subroutine evalout(maxgen,npopsiz,nchrome,nparam,iskip,iend,ibest,fbar,best,&
		& fit_thr,fitness,parent,iparent,g0,g1,ig2)
	!------------------------------------------------------------------------
	! This subroutine evaluates the population, assigns fitness,
	! and establishes the best individual.
	!------------------------------------------------------------------------

		!! subroutine arguments
		integer, intent(in) :: maxgen
		integer, intent(in) :: npopsiz
		integer, intent(in) :: iskip
		integer, intent(in) :: iend
		integer, intent(in) :: nchrome
		integer, intent(in) :: nparam
		real, intent(in)	:: fit_thr
		real, intent(out)   :: fbar
		real, dimension(nparam,npopsiz), intent(in)		:: parent
		integer, dimension(nchrome,npopsiz),intent(in)	:: iparent
		integer, dimension(nchrome), intent(out)		:: ibest
		real, dimension(npopsiz), intent(out)			:: fitness
		real, dimension(nparam),intent(in)				:: g0
        real, dimension(nparam),intent(in)				:: g1
        integer,dimension(nparam),intent(in)			:: ig2


		!! local variables
		integer		:: j,n,k,kk,stat
		integer		:: jbest ! member of the population with the best fitness
		integer,save :: count=1
		real		:: best,prev_best
		real		:: fitsum		!sum of fitness of parents
		real		:: fitsumsq		!sum of square fitness of parents
		real		:: std_dev		!standard deviation for a given generation
		save prev_best

		! initialization
		fitsum=0.0d0
		fitsumsq=0.0d0
		best=100000.0d10

		!! loop over individuals in the population
		do j=1,npopsiz

			!! decode the current chromosome
			call decode(j,nparam,npopsiz,nchrome,parent,iparent,g0,g1,ig2)

			! evaluate fitness of current individual
			call func(npopsiz,nparam,j,fitness(j),parent)

			! update fitness and parameter sums
			fitsum=fitsum+fitness(j)
			fitsumsq=fitsumsq+fitness(j)**2

			! check to see if fitness of this individual is the best fitness
			if (fitness(j) < best) then
				best=fitness(j)
				jbest=j
				ibest(1:nchrome)=iparent(1:nchrome,j)
			end if

		end do

		! compute fitness average and standard deviation for this generation
		fbar=fitsum/dble(npopsiz)
		std_dev=sqrt((dble(npopsiz)*fitsumsq-fitsum**2)/(dble(npopsiz)* &
			& dble(npopsiz-1)))

		! write output information
		write(0,*)'best=',best,' prev_best=',prev_best
		write(0,*)'fbar=',fbar,'std_dev=',std_dev
		write(26,'(f9.6)') best
		write(27,'(f9.6)') fbar
		write(28,'(f9.6)') std_dev
		write(29,'(f9.6)') fbar-best
		

		!! check termination criterion
		if ((best<fit_thr .and.(abs(best-fbar)<1.).and. &
			& (abs(prev_best-best)<0.00001)).or.count==maxgen) then

			!! solution found.
			!! call the fitness function once more just to write the trace out
			call func(npopsiz,nparam,jbest,fitness(j),parent,flag=1)

			write(0,*) 'TERMINATION CRITERIUM SATISFIED. SOLUTION FOUND'
			write(0,*) 'total number of function evaluations =', &
				& (count+1)*npopsiz
			open(unit=31,file='solution',status='unknown',action='write', &
				& iostat=stat)
			if (stat/=0) then
				write(0,*) 'ERROR: could not open file: solution' 
				stop
			else
				do j=1,nparam
					write(31,'f9.4') parent(j,jbest)
				end do
			end if
			close(31)
			stop
		end if

		!! update variables for next generation.
		prev_best=best
		count=count+1

		include 'formats'

		return

	end subroutine evalout
	!----------------------------------------------------------------------

end module ga_fitness
