!! module to implement the niching option for a genetic algorithm

module ga_niche

	implicit none
	
	contains

	!---------------------------------------------------------------------
	subroutine niche (nparam,npopsiz,nchrome, &
		& parent,iparent,fitness,npossibl,nichflg,parmax,parmin,pardel)
	!---------------------------------------------------------------------
	! Implement "niching" through Goldberg's multidimensional phenotypic
	! sharing scheme with a triangular sharing function. To find the
	! multidimensional distance from the best individual, normalize all
	! parameter differences.
	!---------------------------------------------------------------------

		!! subroutine arguments
		integer, intent(in) :: npopsiz
		integer, intent(in) :: nparam
		integer, intent(in) :: nchrome
		integer, dimension(nchrome,npopsiz), intent(in) :: iparent
		integer, dimension(nparam), intent(in)			:: npossibl
		integer, dimension(nparam), intent(in)			:: nichflg
		real, dimension(nparam,npopsiz), intent(in)		:: parent
		real, dimension(npopsiz), intent(inout)			:: fitness
		real, dimension(nparam), intent(in)				:: parmax
		real, dimension(nparam), intent(in)				:: parmin
		real, dimension(nparam), intent(in)				:: pardel

		!! local variables
		real :: alpha,del,del2,sigshar,share,sumshar
		integer :: nniche,j,jj,ii,k

		!
		!	Variable definitions:
		!
		!	alpha   = power law exponent for sharing function; typically = 1.0
		!	del     = normalized multidimensional distance between ii and all
		!				other members of the population (equals the square 
		!				root of del2)
		!	del2    = sum of the squares of the normalized multidimensional
		!				distance between member ii and all other members of 
		!				the population
		!	nniche  = number of niched parameters
		! 	sigshar = normalized distance to be compared with del; in some sense,
		!				1/sigshar can be viewed as the number of regions over 
		!				which the sharing function should focus, e.g. with 
		!				sigshar=0.1, the sharing function will try to clump 
		!				in ten distinct regions of the phase space. A value of 
		!				sigshar on the order of 0.1 seems to work best.
		!	share   = sharing function between individual ii and j
		!	sumshar = sum of the sharing functions for individual ii
		

		alpha=1.0
		sigshar=0.1d0
		nniche=0
		do jj=1,nparam
			nniche=nniche+nichflg(jj)
		end do
		if (nniche == 0) then
			write(0,1900)
			write(24,1900)
			stop
		end if
		do ii=1,npopsiz
			sumshar=0.0d0
			do j=1,npopsiz
				del2=0.0d0
				do k=1,nparam
					if (nichflg(k) > 0) then
						del2=del2+((parent(k,j)-parent(k,ii))/pardel(k))**2
					end if
				end do
				del=(dsqrt(del2))/dble(nniche)
				if (del < sigshar) then
!               	share=1.0-((del/sigshar)**alpha)
					share=1.0d0-(del/sigshar)
				else
					share=0.0d0
				end if
				sumshar=sumshar+share/dble(npopsiz)
			end do
			if (sumshar /= 0.0d0) fitness(ii)=fitness(ii)/sumshar
		end do

		include 'formats'

		return
	end subroutine niche
	!--------------------------------------------------------------------------

end module ga_niche
